using System;
using System.Collections.Generic;
using System.Linq;
using Leap.Unity;
using UnityEngine;
using Object = UnityEngine.Object;
using MacGruber;
using SimpleJSON;
using UnityEngine.Serialization;
using Random = UnityEngine.Random;

namespace CheesyFX
{
    public class Orifice : MonoBehaviour
	{
		private readonly int id;

		public new string name;
		// public List<JSONStorableFloat> bulgeMorphs = new List<JSONStorableFloat>();
		public JSONStorableBool enabledJ = new JSONStorableBool("Enabled", true);

		public JSONStorableFloat sensitivity = new JSONStorableFloat("Sensitivity (ReadMyLips)", 1f, 0f, 5f);
		
		public JSONStorableFloat bulgeScale = new JSONStorableFloat("Bulge Scale", 10f, 0f, 20f);
		public JSONStorableFloat bulgeDepthScale = new JSONStorableFloat("Bulge Depth Scale", 1f, 0f, 4f);
		public JSONStorableFloat bulgeSharpness = new JSONStorableFloat("Bulge Sharpness", 1f, 0f, 5f);
		
		public JSONStorableFloat triggerOffsetUp = new JSONStorableFloat("Trigger Offset Up", 0f, -.05f, .05f);
		public JSONStorableFloat triggerOffsetForward = new JSONStorableFloat("Trigger Offset Forward", 0f, -.05f, .05f);
		
		public JSONStorableFloat depth = new JSONStorableFloat("Depth", 0f, 0f, .4f, true, false);
		public JSONStorableFloat speed = new JSONStorableFloat("Speed", 0f, -5f, 5f, true, false);
		public JSONStorableFloat stretch = new JSONStorableFloat("Stretch", 0f, 0f, 1f, true, false);
		public JSONStorableFloat relaxation;
		
		public JSONStorableFloat depthStimulationThreshold = new JSONStorableFloat("Depth Stimulation Threshold (ReadMyLips)", .10f, 0f, .5f, true);

		private List<CustomFloatTrigger> depthTriggers = new List<CustomFloatTrigger>();
		private List<CustomFloatTrigger> stretchTriggers = new List<CustomFloatTrigger>();
		private List<CustomFloatTrigger> speedTriggers = new List<CustomFloatTrigger>();
		private List<CustomFloatTrigger> relaxationTriggers = new List<CustomFloatTrigger>();
		
		public AudioSource audioSource;

		protected List<object> UIElements = new List<object>();

		public bool isPenetrated{
			get{return _isPenetrated;}
			set
			{
				if (value == _isPenetrated) return;
				_isPenetrated = value;
				if(!value)
				{
					_tip = null;
					depthMeter.transform.parent = null;
					depth.val = 0f;
				}
				ReadMyLips.SetIdleMode();
				// if(name == "Anus") value.Print();
			}
		}

		public Atom penetratingAtom
		{
			set
			{
				FillMeUp.penetratingAtoms[this] = value;
			}
			get { return FillMeUp.penetratingAtoms[this]; }
		}
		private bool _isPenetrated;
		public float lastDepth;
		public Transform enterPointTF;
		public Rigidbody depthReference;
		Collider _tip;
		public GameObject depthMeter;
		public Renderer depthMeterRenderer;
		public GameObject enterTriggerGO;
		public Vector3 enterTriggerBasePosition;
		public Collider enterTriggerCollider;
		[FormerlySerializedAs("triggerHandler")] public OrificeTriggerHandler orificeTriggerHandler;
		public Func<int, float> bulgeFunc;
		public float[] bulgeTargets = new float[8];
		
		private float penetratorWidth;

		public virtual void Init(string name)
		{
			this.name = name;
			AddTrigger();

			CreateDepthMeter();
			
			// bulgeFunc = x =>
			// {
			// 	float exponent = x - (bulgeDepthScale.val * depth.val - 4.4f);
			// 	return bulgeScale.val * penetratorWidth * 4f * Mathf.Pow(1f + bulgeSharpness.val, -exponent * exponent);
			// };
			bulgeFunc = x =>
			{
				float scaledDepth = depth.val * bulgeDepthScale.val * 8f;
				if (scaledDepth < .4f) return 0f;
				
				float exponent = bulgeSharpness.val * (x - scaledDepth-.4f);
				float divisor = 1/(1 + scaledDepth);
				float val = bulgeScale.val * penetratorWidth * (scaledDepth-.4f) * divisor * divisor * Mathf.Pow(2f, -exponent * exponent);
				if (x > scaledDepth) val /= (x - scaledDepth + 1f);
				return val;
			};
			// bulgeScale.setCallbackFunction = x => SetBulgeFunction();
			// bulgeDepthScale.setCallbackFunction = x => SetBulgeFunction();
			// bulgeSharpness.setCallbackFunction = x => SetBulgeFunction();
			
			enabledJ.setCallbackFunction += val => enabled = val;
			triggerOffsetUp.setCallbackFunction = x => UpdateTriggerPosition();
			triggerOffsetForward.setCallbackFunction = x => UpdateTriggerPosition();

			depth.setCallbackFunction += val =>
			{
				for (int i = 0; i < bulgeTargets.Length; i++)
				{
					bulgeTargets[i] = bulgeFunc(i);
				}
			};
		}

		public void AddTrigger()
		{
			var depthTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			depthTrigger.Init(FillMeUp.singleton, depth, $"{name}_{depthTriggers.Count}", false, 0f, 0f, .25f);
			depthTriggers.Add(depthTrigger);
			var stretchTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			stretchTrigger.Init(FillMeUp.singleton, stretch, $"{name}_{stretchTriggers.Count}", false, 0f, 0f, .5f);
			stretchTriggers.Add(stretchTrigger);
			var speedTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			speedTrigger.Init(FillMeUp.singleton, speed, $"{name}_{speedTriggers.Count}", true, 0f, 0f, .5f);
			speedTriggers.Add(speedTrigger);
			if (relaxation != null)
			{
				var relaxationTrigger = gameObject.AddComponent<CustomFloatTrigger>();
				relaxationTrigger.Init(FillMeUp.singleton, relaxation, $"{name}_{relaxationTriggers.Count}", false, 0f, 0f, 5f);
				relaxationTriggers.Add(relaxationTrigger);
			}
		}

		public void OnDisable()
		{
			DoReset();
		}

		public void OnDestroy()
		{
			Destroy(enterTriggerGO);
			Destroy(orificeTriggerHandler);
			Destroy(depthMeter);
			Destroy(audioSource);
		}

		public void OnAtomRename()
		{
			depthTriggers.ForEach(x => x.OnAtomRename());
			stretchTriggers.ForEach(x => x.OnAtomRename());
			speedTriggers.ForEach(x => x.OnAtomRename());
			if (relaxationTriggers != null) relaxationTriggers.ForEach(x => x.OnAtomRename());
		}

		public virtual JSONClass Store(string subScenePrefix, bool storeTriggers = true)
		{
			JSONClass jc = new JSONClass();
			bulgeScale.Store(jc);
			bulgeSharpness.Store(jc);
			bulgeDepthScale.Store(jc);
			sensitivity.Store(jc);
			depthStimulationThreshold.Store(jc);
			if(storeTriggers)
			{
				for (int i = 0; i < depthTriggers.Count; i++)
				{
					jc[depthTriggers[i].name] = depthTriggers[i].Store(subScenePrefix);
				}
				for (int i = 0; i < stretchTriggers.Count; i++)
				{
					jc[stretchTriggers[i].name] = stretchTriggers[i].Store(subScenePrefix);
				}
				for (int i = 0; i < speedTriggers.Count; i++)
				{
					jc[speedTriggers[i].name] = speedTriggers[i].Store(subScenePrefix);
				}
				if (relaxationTriggers != null)
				{
					for (int i = 0; i < relaxationTriggers.Count; i++)
					{
						jc[relaxationTriggers[i].name] = relaxationTriggers[i].Store(subScenePrefix);
					}
				}
			}
			return jc;
		}

		public virtual void Load(JSONClass jc, string subScenePrefix)
		{
			if(jc.HasKey(name))
			{
				JSONClass tc = jc[name].AsObject;
				bulgeScale.Load(tc);
				bulgeSharpness.Load(tc);
				bulgeDepthScale.Load(tc);
				sensitivity.Load(tc);
				depthStimulationThreshold.Load(jc);
				for (int i = 0; i < depthTriggers.Count; i++)
				{
					depthTriggers[i].Load(tc, subScenePrefix);
				}
				for (int i = 0; i < depthTriggers.Count; i++)
				{
					stretchTriggers[i].Load(tc, subScenePrefix);
				}
				for (int i = 0; i < depthTriggers.Count; i++)
				{
					speedTriggers[i].Load(tc, subScenePrefix);
				}
				if (relaxationTriggers != null)
				{
					for (int i = 0; i < relaxationTriggers.Count; i++)
					{
						relaxationTriggers[i].Load(tc, subScenePrefix);
					}
				}
			}
		}

		public void CreateDepthMeter()
		{
			if(depthMeter != null) GameObject.Destroy(depthMeter);
			depthMeter = GameObject.CreatePrimitive(PrimitiveType.Sphere);
			depthMeter.transform.localScale = new Vector3(.005f, .005f, .005f);
			Object.Destroy(depthMeter.GetComponent<Collider>());
			depthMeterRenderer = depthMeter.GetComponent<Renderer>();
			depthMeterRenderer.material.shader = FillMeUp.debugShader;
			switch (name)
			{
				case "Anus":
					depthMeterRenderer.material.color = new Color(1f, 0f, 0.85f, .1f);
					break;
				case "Vagina":
					depthMeterRenderer.material.color = new Color(0f, 0.89f, 1f);
					break;
				default:
					depthMeterRenderer.material.color = new Color(1f, 0.5f, 0f, .2f);
					break;
			}
			depthMeterRenderer.enabled = false;
			depthMeter.tag = "depthMeter";
		}

		void UpdateTriggerPosition()
		{
			enterTriggerGO.transform.localPosition = enterTriggerBasePosition + new Vector3(0f, triggerOffsetUp.val, triggerOffsetForward.val);
		}

		private List<float> lastSpeeds = new List<float>(3);
		public void GetDepthAndSpeed()
		{
			depth.val = Vector3.Distance(depthMeter.transform.position, enterPointTF.position);
			// depth.val += .02f;
			var newSpeed = (depth.val - lastDepth) / Time.fixedDeltaTime;
			
			lastSpeeds.Add(newSpeed);
			if(lastSpeeds.Count > 3) lastSpeeds = lastSpeeds.Skip(1).ToList();
			speed.val = lastSpeeds.Average();
			// speed.val = newSpeed;
			lastDepth = depth.val;
			
		}

		public virtual void GetStretch()
		{
			stretch.val = penetratorWidth;
		}

		// public void SetBulgeMorphs1(){
		// 	float mean = bulgeDepthScale.val*depth.val-3f;
		// 	for(int i=0; i<bulgeMorphs.Count; i++){
		// 		bulgeMorphs[i].val = bulgeScale.val*(float)Math.Pow(1f+bulgeSharpness.val, -(i-mean)*(i-mean));
		// 	}
		// }

		public virtual float Stimulate(float baseStim = 0f, bool changeExpression = false)
		{
			float stimulus = baseStim + 720f * sensitivity.val * (1f + stretch.val) * Mathf.Abs(speed.val) * depth.val * Time.fixedDeltaTime;
			ReadMyLips.Stimulate(stimulus, changeExpression);
			return stimulus;
		}

		private float burstTimer;
		public virtual void FixedUpdate(){
			if(isPenetrated)
			{
				GetDepthAndSpeed();
				GetStretch();
				burstTimer -= Time.fixedDeltaTime;
				Stimulate();
				var moaning = ReadMyLips.PlayMoan();
				if (burstTimer < 0f)
				{
					burstTimer = .5f;
					if (ReadMyLips.randomBurstChance > 0f && ReadMyLips.randomBurstStrenght.val > 0f && Random.Range(0f, 1f) < ReadMyLips.randomBurstChance)
					{
						Stimulate(ReadMyLips.randomBurstStrenght.val, true);
					}
					if(!moaning) ReadMyLips.PlayBurstMoan();
				}
			}
			else
			{
				if (stretch.val > 0f)
				{
					penetratorWidth = 0f;
					speed.val = 0f;
				}
			}
		}

		public virtual void RegisterCollisionWithArbitrary(Collider collider)
		{
			
			_tip = collider;
			depthMeter.transform.position = enterPointTF.position;
			depthMeter.transform.parent = _tip.transform;
			isPenetrated = true;
		}
		
		public virtual void RegisterCollision(Collider collider)
		{
			if (!(collider is CapsuleCollider)) RegisterCollisionWithArbitrary(collider);
			float depth = 0f;
			_tip = collider;
			
			Vector3 enterPoint = enterPointTF.position;
			Vector3 deepestPoint = enterPoint;
			float scale = collider.attachedRigidbody.GetAtom().scaleChangeReceivers[0].scale;
			float radius = 0f;
			foreach(var c in collider.attachedRigidbody.transform.GetComponentsInChildren<Collider>())
			{
				// if (c.gameObject.CompareTag("depthMeter")) continue;
				// c.name.Print();
				var capCol = c as CapsuleCollider;
				if(!ReferenceEquals(capCol, null))
				{
					float capColRadius = capCol.radius * scale;
					if (capColRadius > radius) radius = capColRadius;
					Vector3 farthestPoint = enterPoint;
					Transform transform = capCol.transform;
					int dirIndex = capCol.direction;
					Vector3 direction = new Vector3();
					switch (dirIndex)
					{
						case 0:
							direction = transform.right;
							break;
						case 1:
							direction = transform.up;
							break;
						case 2:
							direction = transform.forward;
							break;
					}

					Vector3[] sphereCenters = 
					{
						transform.TransformPoint(capCol.center + direction * (capCol.height * .5f - capCol.radius)),
						transform.TransformPoint(capCol.center - direction * (capCol.height * .5f - capCol.radius))
					};
					float dist = 0f;
					Vector3 farthestSphereCenter = new Vector3();
					bool anySphereCenterInside = false;
					foreach (var center in sphereCenters)
					{
						Vector3 delta = center - enterPoint;
						if (Vector3.Dot(delta, enterPointTF.up) < 0f) continue;
						// c.name.Print();
						if(!anySphereCenterInside) anySphereCenterInside = true;
						float centerDistance = delta.sqrMagnitude;
						if (centerDistance > dist)
						{
							dist = centerDistance;
							farthestSphereCenter = center;
						}
					}

					if(anySphereCenterInside){
						farthestPoint = farthestSphereCenter +
						                (farthestSphereCenter - enterPoint).normalized * capCol.radius;
						float thisDepth = Vector3.Distance(farthestPoint, enterPoint);
						if (thisDepth > depth)
						{
							depth = thisDepth;
							deepestPoint = farthestPoint;
						}
					}
				}
			}
			depthMeter.transform.position = deepestPoint;
			depthMeter.transform.parent = _tip.transform;
			isPenetrated = true;
			penetratorWidth = radius*8.75f;
			// radius.Print();
		}

		public virtual void DoReset()
		{
			depth.val = stretch.val = penetratorWidth = 0f;
			depthMeter.transform.parent = null;
		}

		public void ClearUI()
		{
			FillMeUp.singleton.RemoveUIElements(UIElements);
		}

		public void SelectTab(int id)
		{
			FillMeUp.singleton.RemoveUIElements(UIElements);
			// lastMenuId = menuItem;
			if (id == 0) CreateTriggersUI();
			else if(id == 1) CreateSettingsUI();
			else if(id == 2) CreateBulgeSettingsUI();
		}

		public virtual void CreateTriggersUI()
		{
			var button = FillMeUp.singleton.CreateButton("Depth Triggers");
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					depthTriggers[0].OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			depth.CreateUI(FillMeUp.singleton, UIElements:UIElements);
			
			button = FillMeUp.singleton.CreateButton("Speed Triggers", true);
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					speedTriggers[0].OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			speed.CreateUI(FillMeUp.singleton, UIElements:UIElements, rightSide:true);
			
			button = FillMeUp.singleton.CreateButton("Stretch Triggers");
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					stretchTriggers[0].OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			stretch.CreateUI(FillMeUp.singleton, UIElements:UIElements);

			if (relaxation != null)
			{
				button = FillMeUp.singleton.CreateButton("Relaxation Triggers", true);
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				button.button.onClick.AddListener(() =>
				{
					{
						FillMeUp.singleton.ClearUI();
						relaxationTriggers[0].OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
					}
				});
				UIElements.Add(button);
				relaxation.CreateUI(FillMeUp.singleton, UIElements:UIElements, rightSide:true);
			}
		}
		
		public virtual void CreateSettingsUI()
		{
			depthStimulationThreshold.CreateUI(FillMeUp.singleton, UIElements: UIElements);
			sensitivity.CreateUI(FillMeUp.singleton, UIElements: UIElements, rightSide:true);
		}

		public virtual void CreateBulgeSettingsUI()
		{
			// bulgeScale.CreateUI(UIElements);
			// bulgeSharpness.CreateUI(UIElements, true);
			// bulgeDepthScale.CreateUI(UIElements);
		}
	}
}