using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using SimpleJSON;
using UnityEngine;
using Random = UnityEngine.Random;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

namespace CheesyFX
{
    public class SlapMe : MVRScript
    {
        public static SlapMe singleton;
        private UnityEventsListener uiListener;
        public static Atom atom;
        private static Transform labia;
        
        public static JSONStorableBool feetSlapsEnabled = new JSONStorableBool("Feet Slaps Enabled", false);
        public static JSONStorableBool selfSlapsEnabled = new JSONStorableBool("Self Slaps Enabled (Hands)", true);
        public static JSONStorableBool slapVocalsEnabled = new JSONStorableBool("Slap Vocal Reaction Enabled", false);
        // public static JSONStorableBool slapMoanEnabled = new JSONStorableBool("Slap Moans Enabled", true);
        // public static JSONStorableBool sexSlapMoanEnabled = new JSONStorableBool("Sex Slap Moans Enabled", true);
        public static JSONStorableBool isSlappingOthers = new JSONStorableBool("Slap Others Enabled", true);
        public static JSONStorableFloat slapVolumeJ = new JSONStorableFloat("Slap Volume", 1f, 0f, 10f, false);
        public static JSONStorableFloat sexSlapVolumeJ = new JSONStorableFloat("Sex Slap Volume", 1f, 0f, 5f, false);
        public static JSONStorableFloat sexSlapThresholdJ = new JSONStorableFloat("Sex Slap Threshold", .5f, .5f, 5f, true);
        public static JSONStorableFloat slapThreshold = new JSONStorableFloat("Slap Threshold", 1f, 1f, 10f, true);
        public static JSONStorableFloat slapStimScale = new JSONStorableFloat("Slap Stimulation Scale", 1f, 0f, 2f, false);
        public static JSONStorableFloat sexSlapStimScale = new JSONStorableFloat("Sex Slap Stimulation Scale", 1f, 0f, 2f, false);

        public static List<Atom> excludedAtoms = new List<Atom>();
        private JSONStorableString excludedInfo = new JSONStorableString("excludeInfo", "<b>Excluded Atoms</b>\n");

        private JSONStorableStringChooser excludedChooser =
	        new JSONStorableStringChooser("atomChooser", null, "", "... this Atom");

        public static float slapVolume;
        public static float sexSlapVolume;
        public static float sexSlapThreshold;

        private static float sexSlapIntensity;
        AudioSource slapAS;
        private static AudioSource sexSlapAS;
        private static IEnumerator sexSlap;
        private static IEnumerator pussySlap;
        
        public static ClipLibrary slapLibrary;
        public static ClipLibrary sexSlapLibrary;
        public static ClipLibrary buttSlapLibrary;
        
        public static string packageUid;

        private GameObject debug;
        public static JSONStorableFloat debugSlider1 = new JSONStorableFloat("debug0", .08f, 0f, 1f, false);
        public static JSONStorableFloat debugSlider2 = new JSONStorableFloat("debug0", .1f, 0f, 1f, false);
        
        public override void Init()
        {
            singleton = this;
            atom = containingAtom;
            packageUid = Utils.GetPackagePath(this);

            slapLibrary = new ClipLibrary("slaps");
            sexSlapLibrary = new ClipLibrary("sexslaps");
            buttSlapLibrary = new ClipLibrary("buttslaps");
            
            var slapsPath = packageUid + "Custom/Scripts/CheesyFX/BodyLanguage/audiobundles/slaps.audiobundle";
            AudioImporter.GetClipsFromAssetBundle(new List<ClipLibrary>{slapLibrary, sexSlapLibrary, buttSlapLibrary},
	            slapsPath);
            
            var labTrigger = atom.rigidbodies.First(x => x.name == "LabiaTrigger").transform;
            labia = new GameObject().transform;
            labia.parent = labTrigger;
            labia.localRotation = Quaternion.identity;
            labia.localPosition = new Vector3(0f, 0f, .03f);

            excludedChooser.choices = SuperController.singleton.GetAtomUIDs();
            SuperController.singleton.onAtomAddedHandlers += AddExclusionChoice;
            SuperController.singleton.onAtomUIDsChangedHandlers += SyncAtomNames;
            slapVolumeJ.AddCallback(val => slapVolume = val * .03f);
            sexSlapVolumeJ.AddCallback(val => sexSlapVolume = val * .6f);
            sexSlapThresholdJ.AddCallback(val => sexSlapThreshold = val);
            
            RegisterBool(feetSlapsEnabled);
            RegisterBool(selfSlapsEnabled);
            RegisterFloat(slapVolumeJ);
            RegisterFloat(sexSlapVolumeJ);
            RegisterFloat(slapThreshold);
            RegisterFloat(sexSlapThresholdJ);
            RegisterFloat(slapStimScale);
            RegisterFloat(sexSlapStimScale);
            RegisterString(excludedInfo);
            excludedInfo.isRestorable = false;
            RegisterStringChooser(excludedChooser);
            ExcludeDefaults();
            UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
            Utils.OnInitUI(CreateUIElement);
            CreateUI();

            // debug = GameObject.CreatePrimitive(PrimitiveType.Sphere);
            // debug.transform.parent = labia;
            // debug.transform.localPosition = Vector3.zero;
            // Destroy(debug.GetComponent<Collider>());
            // debugSlider1.AddCallback(val => debug.transform.localScale = Vector3.one * debugSlider1.val /.5f);
            // debugSlider2.AddCallback(val => debug.transform.localPosition = new Vector3(0f, 0f, val));
        }
        
        public override void InitUI()
        {
	        base.InitUI();
	        if(UITransform == null || uiListener != null) return;
	        uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
	        uiListener.onEnabled.AddListener(() => UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements));
        }

        private void CreateUI()
        {
	        slapVolumeJ.CreateUI(script:this);
	        sexSlapVolumeJ.CreateUI(script:this, rightSide:true);
	        slapThreshold.CreateUI();
	        sexSlapThresholdJ.CreateUI(rightSide:true);
	        slapStimScale.CreateUI();
	        sexSlapStimScale.CreateUI(rightSide:true);
	        
	        // UIManager.SetupInfoOneLine(null, "Expressions", false, false, this);
	        // var textfield = CreateTextField(new JSONStorableString("", "Expressions"));
	        // textfield.height = 50f;
	        // textfield.UItext.alignment = TextAnchor.MiddleCenter;
	        
	        var button = CreateButton("Disable Touches with...");
	        button.button.onClick.AddListener(() => Exclude(excludedChooser.val, true));
	        button = CreateButton("Enable Touches with...");
	        button.button.onClick.AddListener(() => Exclude(excludedChooser.val, false));
	        excludedChooser.CreateUI(script: this, rightSide: true,chooserType:2);
	        button = CreateButton("Reset");
	        button.button.onClick.AddListener(ResetExcluded);
	        var textfield = CreateTextField(excludedInfo, true);
	        textfield.height = 400f;
	        selfSlapsEnabled.CreateUI(this);
	        feetSlapsEnabled.CreateUI(this);
	        // debugSlider1.CreateUI();
	        // debugSlider2.CreateUI();
        }
        
        public static void Slap(TouchZone touchZone, Collision collision)
        {
	        if (!feetSlapsEnabled.val && touchZone.parents.Any(x => x.name == "Feet")) return;
			ContactPoint contactPoint = collision.contacts[0];
			Vector3 velocity = collision.relativeVelocity;
			float intensity = Math.Abs(Vector3.Dot(velocity, contactPoint.normal));
			if (intensity < sexSlapThreshold) return;
			AudioSource audioSource = touchZone.slapAudioSource;
			audioSource.transform.position = contactPoint.point;
			Atom collidingAtom = collision.rigidbody.GetAtom();
			bool selfSlap = collidingAtom == atom;
			bool slapsOtherPerson = false;
			bool slappedByPerson = false;
			bool slappedByPersonOrToys = collidingAtom.type == "Dildo" || collidingAtom.type == "Paddle" || collidingAtom.type == "Capsule";;
			string collidingRegion = collision.rigidbody.GetRegionName();
			if(collidingRegion != null)
			{
				bool contactWithPerson = collidingAtom.type == "Person";
				slapsOtherPerson = contactWithPerson && touchZone.parents.Any(x => x.name.Contains("Hand"));
				bool slappedByHand = collidingRegion.Contains("Hand") || collidingRegion.Contains("Thumb") || collidingRegion.Contains("Index") || collidingRegion.Contains("Mid") || collidingRegion.Contains("Ring") || collidingRegion.Contains("Pinky");
				
				slappedByPerson = contactWithPerson && (slappedByHand || collidingRegion.Contains("Arm"));
				slappedByPersonOrToys = slappedByPersonOrToys || slappedByPerson;
				                       
				

				// touchZone.name.Print();
				
			}
			bool collidingAtomIsPenetrating = FillMeUp.penetratingAtoms.Values.ToList().Contains(collidingAtom);
			if (collidingAtomIsPenetrating &&
			    (collidingRegion == "Penis" || collidingRegion == "DildoPenis")) return;
			if (
				collidingAtomIsPenetrating
				&& collidingRegion != null
				&& (collidingRegion == "Pelvis" || collidingRegion == "Abdomen" ||
				    collidingRegion.Contains("Thigh") || collidingRegion == "Testes" ||
				    collidingRegion == "DildoSack"))
			{
				if ((touchZone.name.Contains("Thigh") || touchZone.name.Contains("Glutes") ||
				     touchZone.name == "Labia" || touchZone.name == "Anus" || touchZone.name == "Vagina"))
				{
					if (intensity < sexSlapIntensity) return;
					sexSlapIntensity = intensity;
					sexSlapAS = audioSource;
					if (sexSlap == null) sexSlap = SexSlap().Start();

				}
			}
			else if(Vector3.Distance(labia.position, contactPoint.point) < .08f)
			{
				if(intensity < sexSlapIntensity) return;
				sexSlapIntensity = intensity;
				sexSlapAS = audioSource;
				if (pussySlap == null) pussySlap = PussySlap().Start();
			}
			else{
				if(intensity < slapThreshold.val) return;
				float direction = intensity / velocity.magnitude;
				// (collision.rigidbody.name+"->"+touchZone.name).Print();
				List<BodyRegion> touchedParents = touchZone.parents;
				if (touchedParents.Any(x => x.name.Contains("Hand")) ||
				    touchedParents.Any(x => x.name.Contains("ForeArm")))
				{
					if (!slapsOtherPerson || !isSlappingOthers.val) return;
				}
				bool slappedByOtherPerson = !selfSlap && slappedByPersonOrToys;
				// touchZone.name.Print();
				touchZone.Slap(intensity, direction, touchZone,slapVocalsEnabled.val && slappedByOtherPerson, slappedByOtherPerson);
			}
		}
        
        private static IEnumerator PussySlap(){
	        AudioClip clip = buttSlapLibrary.GetRandomClip();
	        yield return new WaitForSeconds(.05f);
	        float volumeFactor = sexSlapIntensity*sexSlapIntensity*sexSlapVolumeJ.val*1.3f;
	        sexSlapAS.pitch = Mathf.Lerp(1.2f, 1.5f, sexSlapIntensity);
	        // volumeFactor.Print();
	        // sexSlapIntensity.Print();
	        sexSlapAS.PlayOneShot(clip, volumeFactor);
	        var vagAS = FillMeUp.vagina.audioSource;
	        if (vagAS == null)
	        {
		        sexSlapIntensity = 0f;
		        pussySlap = null;
		        yield break;
	        }
	        vagAS.volume = Mathf.Lerp(.75f, 1f,sexSlapIntensity);
	        
	        vagAS.PlayOneShot(FillMeUp.squishLibrary.GetRandomClip());
	        
	        // if(globalSlapMoanReactionEnabled.val) AudioManager.moanClips.PlayRandom(allowRepeat:true);
	        if(ReadMyLips.singleton.enabled) ReadMyLips.Stimulate(sexSlapStimScale.val * sexSlapIntensity * 8f);
	        ReadMyLips.PlaySlapMoan(true);
	        sexSlapIntensity = 0f;
	        pussySlap = null;
        }
        
        private static IEnumerator SexSlap(){
	        AudioClip clip = buttSlapLibrary.GetRandomClip();
	        yield return new WaitForSeconds(.05f);
	        float delta = sexSlapIntensity - sexSlapThreshold;
	        float volumeFactor = delta * delta * sexSlapVolume;
	        sexSlapAS.pitch = Mathf.Lerp(.6f, 1.2f, sexSlapIntensity);
	        sexSlapAS.PlayOneShot(clip, volumeFactor);
	        if(ReadMyLips.singleton.enabled) ReadMyLips.Stimulate(sexSlapStimScale.val * sexSlapIntensity * 6f);
	        ReadMyLips.PlaySlapMoan(true);
	        sexSlapIntensity = 0f;
	        sexSlap = null;
        }

        private void Exclude(string uid, bool val)
        {
	        Atom atom = SuperController.singleton.GetAtomByUid(uid);
	        if (atom == null) return;
	        if (val)
	        {
		        if(!excludedAtoms.Contains(atom))
		        {
			        excludedAtoms.Add(atom);
			        excludedInfo.val += atom.uid + "\n";
		        }
	        }
	        else if(excludedAtoms.Remove(atom))
	        { 
		        excludedInfo.val = "<b>Excluded Atoms</b>\n"+string.Join("\n", excludedAtoms.Select(x => x.uid).ToArray());
	        }
        }
        
        private void AddExclusionChoice(Atom atom)
        {
	        if (excludedChooser.choices.Contains(atom.uid)) return;
	        var choices = excludedChooser.choices;
	        choices.Insert(0, atom.uid);
	        excludedChooser.choices = new List<string>(choices);
        }

        private void ExcludeDefaults()
        {
	        string[] typesToExclude = {"Panel", "Slate", "Wall", "Glass", "CustomUnityAsset", "Apt", "Cyberpunk", "Deco"};
	        foreach (var atom in SuperController.singleton.GetAtoms())
	        {
		        if(typesToExclude.Any(x => atom.type.Contains(x))) Exclude(atom.uid, true);
	        }
        }
        
        private void SyncAtomNames(List<string> uids)
        {
	        excludedChooser.choices = new List<string>(uids);
	        var ignoredUids = excludedAtoms.Select(x => x.uid).ToArray();
	        excludedInfo.val = "<b>Excluded Atoms</b>\n"+string.Join("\n", ignoredUids);
        }
        
        private void ResetExcluded()
        {
	        excludedAtoms.Clear();
	        excludedInfo.val = "<b>Excluded Atoms</b>\n";
        }

        private void OnDestroy()
        {
	        foreach(TouchZone touchZone in BodyRegionMapping.touchZones.Values.ToList()){
		        touchZone.Destroy();
	        }
            AudioImporter.UnloadBundles();
            SuperController.singleton.onAtomAddedHandlers -= AddExclusionChoice;
            SuperController.singleton.onAtomUIDsChangedHandlers -= SyncAtomNames;
            ReadMyLips.Reset();
            sexSlap.Stop();
            sexSlap = null;
            pussySlap.Stop();
            pussySlap = null;
            Destroy(labia.gameObject);
        }
        
        public override void LateRestoreFromJSON(
	        JSONClass jc,
	        bool restorePhysical,
	        bool restoreAppearance,
	        bool setMissingToDefault
        )
        {
	        if (jc.HasKey("excludeInfo"))
	        {
		        ResetExcluded();
		        var uidsToIgnore = ((string)jc["excludeInfo"]).Split(new [] {"\n"}, StringSplitOptions.None).Skip(1);
		        foreach (var uid in uidsToIgnore)
		        {
			        Exclude(uid, true);
		        }
	        }
        }
    }
}