using System.Collections.Generic;
using UnityEngine;

namespace CheesyFX
{
    public static class BodyRegionMapping
    {
	    public static List<string> touchZoneNames = new List<string>{
		    "Abdomen", "Pelvis",
		    "lGlutes","rGlutes","lThigh","rThigh","lShin","rShin",
		    "Chest","lBreast","rBreast","lNipple","rNipple",  
		    "Head", "Eyes", "Lips", "Mouth", "Throat",
			
		    "CustomVag", "Labia", "Vagina","DeepVagina","DeeperVagina","Anus",
			
		    "lFoot","lBigToe", "lSmallToe1", "lSmallToe2", "lSmallToe3", "lSmallToe4",
		    "rFoot","rBigToe", "rSmallToe1", "rSmallToe2", "rSmallToe3", "rSmallToe4",
		    "lShoulder","rShoulder","lForeArm","rForeArm",
		    "lPalm","lThumb","lIndex","lMid","lRing","lPinky",
		    "rPalm","rThumb","rIndex","rMid","rRing","rPinky",
	    };
	    public static List<string> bodyRegionNames = new List<string>{
		    "Legs", "lLeg", "rLeg", "Thighs", "Shins",
		    "Crotch", "Glutes", "Breasts", "Nipples",
		    "Feet", "Toes", "lToes", "rToes",
		    "Arms", "lArm", "rArm", "Shoulders", "ForeArms",
		    "lHand", "rHand", "Hands","lFingers", "rFingers", "Fingers",
	    };
	    public static List<string> cumRegionNames = new List<string>{
		    "lLeg", "rLeg", "Thighs", "Shins",
		    "Crotch", "Glutes", "lBreast", "rBeast", "Abdomen", "Pelvis",
		    "Feet", "Toes", "lToes", "rToes",
		    // "lArm", "rArm", "Shoulders", "ForeArms",
		    "lHand", "rHand",
		    "Eyes", "Lips", "Mouth"
	    };
	    public static List<string> topRegionNames = new List<string>{
		    "Legs", "Crotch", "Glutes", "Breasts", "Feet", "Arms", "Hands", "Head"
	    };
	    
	    public static Dictionary<string, TouchZone> touchZones = new Dictionary<string, TouchZone>();
	    public static Dictionary<string, BodyRegion> bodyRegions = new Dictionary<string, BodyRegion>();
	    public static Dictionary<string, BodyRegion> regionsByLowerName;
	    public static List<BodyRegion> topRegions = new List<BodyRegion>();
	    
	    public static Dictionary<string, string> bodyRegionByRigidbodyName = new Dictionary<string, string>();

	    public static void Init(Atom atom)
	    {
		    foreach (var rb in atom.rigidbodies)
		    {
			    bodyRegionByRigidbodyName[rb.name] = MapToRegion(rb);
		    }
		    MapMiscRegions();
		    
		    foreach(string name in bodyRegionNames){
			    BodyRegion region = new BodyRegion(name);
			    bodyRegions[name] = region;
			    if (topRegionNames.Contains(name))
			    {
				    topRegions.Add(region);
			    }
		    }
		    foreach(string name in touchZoneNames){
			    TouchZone touchZone = new TouchZone(name);
			    touchZones[name] = touchZone;
			    if (topRegionNames.Contains(name)) topRegions.Add(touchZone);
			    // touchZone.GetCumRegion();
		    }
		    BuildHierarchy();
	    }

	    public static string GetRegionName(string rbName)
	    {
		    string region;
		    bodyRegionByRigidbodyName.TryGetValue(rbName, out region);
		    return region;
	    }

	    public static string GetRegionName(this Rigidbody rb)
	    {
		    return GetRegionName(rb.name);
	    }
	    
	    public static TouchZone GetRegion(this Rigidbody rb)
	    {
		    TouchZone tz;
		    touchZones.TryGetValue(rb.GetRegionName(), out tz);
		    return tz;
	    }
	    
        public static string MapToRegion(Rigidbody rb)
		{
			string name = rb.name;
			string output = null;
			if (name.Contains("Link") || name.Contains("Control")) return null;

			if (name.Contains("Lip")) output = "Lips";
			else if (name.Contains("FaceHardLeft5") || name.Contains("FaceHardRight5"))
			{
				output = "Eyes";
			}
			else if (name.Contains("head")) output = "Head";
			else if (name.Contains("chest")) output = "Chest";
			else if (name.Contains("pelvis")) output = "Pelvis";
			else if (name.Contains("rThigh")) output = "rThigh";
			else if (name.Contains("lThigh")) output = "lThigh";
			else if (name.Contains("rShin")) output = "rShin";
			else if (name.Contains("lShin")) output = "lShin";
			else if (name.Contains("rFoot")) output = "rFoot";
			else if (name.Contains("rBigToe")) output = "rBigToe";
			else if (name.Contains("rSmallToe1")) output = "rSmallToe1";
			else if (name.Contains("rSmallToe2")) output = "rSmallToe2";
			else if (name.Contains("rSmallToe3")) output = "rSmallToe3";
			else if (name.Contains("rSmallToe4")) output = "rSmallToe4";
			else if (name.Contains("lFoot")) output = "lFoot";
			else if (name.Contains("lBigToe")) output = "lBigToe";
			else if (name.Contains("lSmallToe1")) output = "lSmallToe1";
			else if (name.Contains("lSmallToe2")) output = "lSmallToe2";
			else if (name.Contains("lSmallToe3")) output = "lSmallToe3";
			else if (name.Contains("lSmallToe4")) output = "lSmallToe4";
			else if (name.Contains("PhysicsMeshJointleft glute")) output = "lGlutes";
			else if (name.Contains("PhysicsMeshJointright glute")) output = "rGlutes";
			else if (name.Contains("AutoColliderFemaleAutoColliderslPectoral")) output = "lBreast";
			else if (name.Contains("AutoColliderFemaleAutoCollidersrPectoral")) output = "rBreast";
			else if (name.Contains("PhysicsMeshJointleft")) output = "lBreast";
			else if (name.Contains("PhysicsMeshJointright")) output = "rBreast";
			else if (name=="lNipple") output = "lNipple";
			else if (name=="rNipple") output = "rNipple";
			else if (name.Contains("JointG")) output = "Labia";
			else if (name.Contains("PhysicsMeshJointgen")) output = "Labia";
			else if (name.Contains("PhysicsMeshJointlab")) output = "Labia";
			else if (name.Contains("JointA")) output = "Anus";
			else if (name.Contains("PhysicsMeshJointan")) output = "Anus";
			else if (name.Contains("PhysicsMeshJointbridge")) output = "Anus";
			else if (name.Contains("PhysicsMeshJoint")) output = "Lips";
			else if (name.Contains("chest")) output = "Chest";
			else if (name.Contains("abdomen")) output = "Abdomen";
			else if (name.Contains("pelvis")) output = "Pelvis";
			else if (name.Contains("lForeArm") || name.Contains("lShldr")) output = "lArm";
			else if (name.Contains("rForeArm") || name.Contains("rShldr")) output = "rArm";
			else if (name.Contains("lHand") || name.Contains("lCarpal")) output = "lPalm";
			else if (name.Contains("lThumb")) output = "lThumb";
			else if (name.Contains("lIndex")) output = "lIndex";
			else if (name.Contains("lMid")) output = "lMid";
			else if (name.Contains("lRing")) output = "lRing";
			else if (name.Contains("lPinky")) output = "lPinky";
			else if (name.Contains("rHand") || name.Contains("rCarpal")) output = "rPalm";
			else if (name.Contains("rThumb")) output = "rThumb";
			else if (name.Contains("rIndex")) output = "rIndex";
			else if (name.Contains("rMid")) output = "rMid";
			else if (name.Contains("rRing")) output = "rRing";
			else if (name.Contains("rPinky")) output = "rPinky";
			
			// else if (name.Contains("Gen")) output = "Penis";
			// else if (name == "Testes") output = "Testes";
			// else if (name == "sack") output = "DildoSack";
			// else if (name == "b1" || name == "b2" || name == "b3") output = "DildoPenis";
			// else if (name == "object") output = "Object";

			// if (name=="lNipple")
			// {
			// 	rb.name.Print();
			// 	rb.GetComponents<Collider>().ToList().ForEach(x => x.isTrigger.Print());
			// }
			return output;
		}

        private static void MapMiscRegions()
        {
	        bodyRegionByRigidbodyName["sack"] = "DildoSack";
	        bodyRegionByRigidbodyName["b1"] = "DildoPenis";
	        bodyRegionByRigidbodyName["b2"] = "DildoPenis";
	        bodyRegionByRigidbodyName["b3"] = "DildoPenis";
	        bodyRegionByRigidbodyName["Gen1"] = "Penis";
	        bodyRegionByRigidbodyName["Gen2"] = "Penis";
	        bodyRegionByRigidbodyName["Gen3"] = "Penis";
	        bodyRegionByRigidbodyName["Testes"] = "Testes";
        }
        
        public static void BuildHierarchy(){
			touchZones["Anus"].parent = bodyRegions["Crotch"];
			touchZones["Labia"].parent = bodyRegions["Crotch"];
			touchZones["Vagina"].parent = bodyRegions["Crotch"];
			touchZones["DeepVagina"].parent = touchZones["Vagina"];
			touchZones["DeeperVagina"].parent = touchZones["DeepVagina"];

			foreach(string s in new []{"l", "r"}){
				touchZones[s+"Thigh"].parent = bodyRegions["Thighs"];
				touchZones[s+"Shin"].parent = bodyRegions["Shins"];

				touchZones[s+"Shoulder"].parent = bodyRegions["Shoulders"];
				touchZones[s+"ForeArm"].parent = bodyRegions["ForeArms"];

				touchZones[s+"Glutes"].parent = bodyRegions["Glutes"];

				touchZones[s+"Breast"].parent = bodyRegions["Breasts"];
				touchZones[s+"Nipple"].parent = bodyRegions["Nipples"];

				foreach(string finger in new []{"Thumb","Index","Mid","Ring","Pinky"})
				{
					touchZones[s + finger].parent = bodyRegions[s+"Fingers"];
				}
				bodyRegions[s+"Hand"].parent = bodyRegions["Hands"];
				touchZones[s+"Palm"].parent = bodyRegions[s+"Hand"];
				bodyRegions[s+"Fingers"].parent = bodyRegions[s+"Hand"];

				
				foreach(string toe in new string[]{"BigToe","SmallToe1","SmallToe2","SmallToe3","SmallToe4"}){
					touchZones[s+toe].parent = bodyRegions[s+"Toes"];
				}
				bodyRegions[s+"Toes"].parent = touchZones[s+"Foot"];
				touchZones[s+"Foot"].parent = bodyRegions["Feet"];
			}
			bodyRegions["Thighs"].parent = bodyRegions["Legs"];
			bodyRegions["Shins"].parent = bodyRegions["Legs"];
			
			bodyRegions["Shoulders"].parent = bodyRegions["Arms"];
			bodyRegions["ForeArms"].parent = bodyRegions["Arms"];

			bodyRegions["Nipples"].parent = bodyRegions["Breasts"];
			// foreach(BodyRegion bodyRegion in allRegions){
			// 	BodyRegion parent = bodyRegion.parent;
			// 	int i=0;
			// 	while(parent != null && i<10){
			// 		parent.children.Add(bodyRegion);
			// 		parent = parent.parent;
			// 		i+=1;
			// 	}
			// 	// if(bodyRegion.parent != null) bodyRegion.parent.children.Add(bodyRegion);
			// }
			foreach(TouchZone touchZone in touchZones.Values.ToList()){
				touchZone.GetParents();
			}
			// bodyRegions["Legs"].children.ForEach(x => x.name.Print());
		}
    }
}