using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace CheesyFX
{
    public class TouchCollisionListener
	{
		public TouchZone touchZone;
		public Collision lastCollision;
		public JSONStorableBool touchEnabled = new JSONStorableBool("Touch Enabled", true);
		
		bool hasTriggers;
		public List<TouchCollisionTrigger> collisionTriggers = new List<TouchCollisionTrigger>();
		// public Dictionary<CollisionEvent, JSONStorableFloat> events = new Dictionary<CollisionEvent, JSONStorableFloat>();
		public HashSet<Rigidbody> stuffColliding = new HashSet<Rigidbody>();
		public bool isOnStay;
		public float timeTouched;
		float onStayTimeout;
		IEnumerator onStayRoutine;
		public JSONStorableFloat onStayTolerance = new JSONStorableFloat("OnStay Tolerance", .1f, .1f, 2f, false, true);
		public JSONStorableFloat reactionChance = new JSONStorableFloat("Reaction Chance", 1f, 0f, 1f, true, true);

		public JSONStorableStringChooser addEventChoice;
		public JSONStorableFloat SlapIntensityFactor = new JSONStorableFloat("SlapIntensityFactor", 1f, 0f,5f);
		public UnityEvent onEnterEvent = new UnityEvent();
		public UnityEvent onExitEvent = new UnityEvent();
		

		public bool slapVocalsEnabled = true;

		public TouchCollisionListener(TouchZone touchZone){
			this.touchZone = touchZone;
			InitTriggers();
			// slapEnabled = new JSONStorableBool("Slapping Enabled", true, ToggleSlapAudioObject);
			// ToggleSlapAudioObject(slapEnabled.defaultVal);

		}

		// public void AddEvent(CollisionEvent colEvent){
		// 	events[colEvent] = new JSONStorableFloat("onStayTimer", 0f, 0f, 100f);
		// 	// if(!hasTriggers && touchEnabled.val){
		// 	// 	// foreach(Rigidbody rb in bodyRegion.bodyRegionManager.atom.GetBodyRegion(bodyRegion.name)){
		// 	// 	// 	AddCollisionTrigger(bodyRegion, rb);
		// 	// 	// }
		// 		
		// 	// }
		// }
		//
		// public void RemoveEvent(CollisionEvent colEvent){
		// 	events.Remove(colEvent);
		// 	// if(events.Count == 0){
		// 	// 	DestroyTriggers();
		// 	// 	// collisionTriggers.ForEach(x => MonoBehaviour.Destroy(x));
		// 	// }
		// }
		//
		// public void SetupTriggers(bool enabled){
		// 	// events.Count.Print();
		// 	if(enabled && events.Count > 0){
		// 		// bodyRegion.name.Print();
		// 		foreach(Rigidbody rb in touchZone.bodyManager.atom.GetRigidbodiesByRegion(touchZone.name)){
		// 			// rb.name.Print();
		// 			AddCollisionTrigger(touchZone, rb);
		// 			// if(bodyRegion.name == "rHand") rb.name.Print();
		// 		}
		// 		hasTriggers = true;
		// 	}
		// 	else{
		// 		if(hasTriggers) Destroy();
		// 	}
		// }

		void AddCollisionTrigger(TouchZone bodyRegion, Rigidbody rb){
			if(rb != null){
				TouchCollisionTrigger collisionTrigger = rb.gameObject.GetComponent<TouchCollisionTrigger>();
				if(collisionTrigger == null){
					collisionTrigger = rb.gameObject.AddComponent<TouchCollisionTrigger>();
					collisionTrigger.Init(bodyRegion, rb);
				}
				collisionTriggers.Add(collisionTrigger);
			}
		}

		public void InitTriggers(){
			foreach(Rigidbody rb in FillMeUp.atom.GetRigidbodiesByRegion(touchZone.name)){
				AddCollisionTrigger(touchZone, rb);
			}
		}


		public void Destroy(){
			collisionTriggers.ForEach(x => Object.Destroy(x));
			collisionTriggers.Clear();
			onStayRoutine.Stop();
			hasTriggers = false;
		}

		public void RegisterCollision(TouchCollisionTrigger collisionTrigger){
			if(touchEnabled.val){
				if(!stuffColliding.IsSupersetOf(collisionTrigger.stuffColliding)){
					lastCollision = collisionTrigger.lastCollision;
					SlapMe.Slap(touchZone, lastCollision);
					// if(isOnStay) events.Keys.ToList().ForEach(x => x.Trigger(touchZone, lastCollision, 0));
					stuffColliding.UnionWith(collisionTrigger.stuffColliding);
				}
				onStayTimeout = onStayTolerance.val;
				if(!isOnStay)
				{
					isOnStay = true;
					// events.Keys.ToList().ForEach(x => x.Trigger(touchZone, lastCollision, 0));
					stuffColliding.UnionWith(collisionTrigger.stuffColliding);
					onStayRoutine = OnStayRoutine();
					onStayRoutine.Start();
					onEnterEvent.Invoke();
				}
			}
		}

		public void DeregisterCollision(Collision collision){
			stuffColliding.Remove(collision.rigidbody);
			if(stuffColliding.Count == 0) onStayTimeout = onStayTolerance.val;
		}
		
		public void DeregisterCollision(Collider collider){
			stuffColliding.Remove(collider.attachedRigidbody);
			if(stuffColliding.Count == 0) onStayTimeout = onStayTolerance.val;
		}

		public IEnumerator OnStayRoutine(){
			while(stuffColliding.Count > 0 || onStayTimeout > 0f)
			{
				// touchZone.bodyManager.arousal.arousal.val += touchZone.topParent.touchArousalRate.val;
				if(stuffColliding.Count == 0) onStayTimeout -= Time.deltaTime;
				// foreach(var kvp in events){
				// 	if(kvp.Key.type.Contains(1)){
				// 		if(kvp.Key.onStayPeriod_mean.val > 0f){
				// 			kvp.Value.val -= Time.deltaTime;
				// 			if(kvp.Value.val <= 0f){
				// 				kvp.Value.val = NormalDistribution.GetValue(kvp.Key.onStayPeriod_mean.val, kvp.Key.onStayPeriod_delta.val);
				// 				if(ShouldReact()) kvp.Key.Trigger(touchZone, lastCollision, 1);
				// 			}
				// 		}
				// 		else{
				// 			if(ShouldReact()) kvp.Key.Trigger(touchZone, lastCollision, 1);
				// 		}
				// 	}
				// }
				yield return new WaitForSeconds(Time.deltaTime);
			}
			// events.Keys.ToList().ForEach(x => x.Trigger(touchZone, lastCollision, 2));
			isOnStay = false;
			onExitEvent.Invoke();
		}

		public bool ShouldReact(){
			float rand = UnityEngine.Random.Range(0f, 1f);
			return rand < reactionChance.val;
		}

		// public void RegisterLookAt(){
		// 	BodyRegion region = BodyManager.detailedViewScan.val ? touchZone : touchZone.topParent;
		// 	if (region.numLookAtColliders == 0)
		// 	{
		// 		BodyManager.regionsLookedAt.Add(region);
		// 		// region.parents.ForEach(x => BodyManager.regionsLookedAt.Add(x));
		// 	}
		// 	region.numLookAtColliders += 1;
		// }
		//
		// public void DeregisterLookAt()
		// {
		// 	BodyRegion region = BodyManager.detailedViewScan.val ? touchZone : touchZone.topParent;
		// 	if (region.numLookAtColliders > 0)
		// 	{
		// 		region.numLookAtColliders -= 1;
		// 		if (region.numLookAtColliders == 0)
		// 		{
		// 			BodyManager.regionsLookedAt.Remove(region);
		// 			// region.parents.ForEach(x => BodyManager.regionsLookedAt.Remove(x));
		// 		}
		// 	}
		// }

		public void CreateUI(){
			// UIManager.ClearItems();
			// UIManager.CreateItemsRemove(onEnterEvents.ToArray(), RemoveEvent, tabLevel);
			// UIManager.CreateItemsRemove(onStayEvents.ToArray(), RemoveEvent, tabLevel);
			// UIManager.CreateItemsRemove(onExitEvents.ToArray(), RemoveEvent, tabLevel);
			// UIManager.CreateItemsToggle(bodyRegion.bodyRegionManager.collisionEvents.ToArray(), ToggleEvent, 1);
		}
	}
}