using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Object = UnityEngine.Object;
using Random = UnityEngine.Random;

namespace CheesyFX
{
    public class TouchZone : BodyRegion
    {
        public TouchCollisionListener touchCollisionListener;
        // public CumRegion cumRegion;
        public List<Atom> atomsToCollideWith = new List<Atom>();
        public List<Action> onWatchActions = new List<Action>();
        
        public JSONStorableBool slapEnabled;
        public GameObject slapAudioObject;
        public AudioSource slapAudioSource;
        public float slapIntensity;
        public float slapDirection;
        public IEnumerator slap;
        private System.Random rdn = new System.Random();

        public TouchZone(string region) : base(region){
            touchCollisionListener = new TouchCollisionListener(this);
            ToggleSlapAudioObject(true);
        }
        
        public void ToggleSlapAudioObject(bool enabled){
            if(enabled){
                slapAudioObject = new GameObject();
                slapAudioSource = slapAudioObject.AddComponent<AudioSource>();
                slapAudioSource.spatialBlend = 1f;
                slapAudioSource.spatialize = true;
            }
            else Object.Destroy(slapAudioObject);
        }

        public void Slap(float intensity, float direction, TouchZone touchZone, bool doVocals, bool doMoan)
        {
            if(intensity <= slapIntensity) return;
            slapIntensity = intensity;
            slapDirection = direction;
            if (slap == null) slap = DeferrredSlap(touchZone, doVocals, doMoan).Start();
        }
        
        private IEnumerator DeferrredSlap(TouchZone touchZone, bool doVacals, bool doMoan)
        {
            // if (!slapVocalsEnabled) yield break;
            AudioClip clip = SlapMe.slapLibrary.GetRandomClip();

            yield return new WaitForSeconds(.05f);
            float delta = slapIntensity - SlapMe.slapThreshold.val;
            float volumeFactor = delta * delta * SlapMe.slapVolume;
            slapAudioSource.pitch = Mathf.Lerp(.9f, 1.5f* slapPitchFactor.val, slapIntensity*.166f * (1f-slapDirection * slapDirection));
            // volumeFactor.Print();
            slapAudioSource.PlayOneShot(clip, volumeFactor);
            if (ReadMyLips.blinkTimeout <= 0f && Random.Range(0f, slapIntensity) > 3f)
            {
                ReadMyLips.eyelidBehavior.Blink();
                ReadMyLips.blinkTimeout = 1f;
            }
            if(ReadMyLips.singleton.enabled) ReadMyLips.Stimulate(SlapMe.slapStimScale.val * slapIntensity, doStim:true);
            if(doMoan && slapIntensity > SlapMe.slapThreshold.val * 1.5f) ReadMyLips.PlaySlapMoan();
            // var prompt = "{{user}} slaps {{char}}'s "+ touchZone.name +" with a " + collision.rigidbody.GetAtom().type+
            //     " with an intensity of " + slapIntensity.ToString();
            // prompt.Print();
            // touchZone.name.Print();
            slapIntensity = 0f;
            slap = null;
        }

        public void Destroy()
        {
            touchCollisionListener.Destroy();
            Object.Destroy(slapAudioObject);
            slap.Stop();
            slap = null;
        }

        // public void GetCumRegion()
        // {
        //     foreach (var prnt in parents)
        //     {
        //         if (bodyManager.cumRegions.TryGetValue(prnt.name, out cumRegion))
        //         {
        //             // prnt.name.Print();
        //             return;
        //         }
        //     }
        // }
    }
}